#!/usr/bin/python
'''Usage: mythalsig.py [OPTIONS]
MythTV Automatic Lyng-Sat Icon Grabber v2
For fuzzy matching channels in MythTV database and grabbing icons for them.
Options:
    -h, --help              Print this info.
    -r, --regions           Comma seperated region list i.e. us-il,us-ma,us.
                            (Default: uk)
    -s, --source=SOURCE     Only grab icons for channels provided by SOURCE ID.
                            (Default: All sources)
    -o, --overwrite         Create all new data for regions (This will destroy
                            old settings!)
    -c, --callsigns         Work in callsign mode (do not d/l icons, let mythtv.
                            do it using iconmap tables)
    -f, --force-update      Override the built in caching of the web source.

Steps:
    1. Grab list of stations from db
    2. Grab list of icons available from lyng-sat
    3. For each station, do a fuzzy match
    If in callsign mode
        4. Update the tables in db
           (You then run "mythfilldatabase --update-icon-map")
    else
        4. Download icons
        5. Insert full path to icon into db
           (You then move the dowloaded icons to the appropriate directory)
    6. Update the cache and save to disk
    
Note: Lots of places don't use callsigns, hence they are optional.
'''
# # # # # # User Settings Here # # # # # #
# database settings
db_host='192.168.72.3'
db_port=3306
db_name='mythconverg'
db_user='mythtv'
db_passwd='mythtv'
# icons are downloaded here (might not be running script on mythtv system)
icon_dir_local="icons"
# path that icons will be located in on mythtv system (manually copy logos to this path, this is what is placed in the db)
icon_dir_mythtv='/mnt/store/mythtv/.mythtv/logos'
# # # # # # End of User Settings # # # # # #

# # # # # # Program Settings Here # # # # # #
# You probably don't want to change these
# pages at lyngsat that contain the icons
lyngsat_sub_page='http://www.lyngsat-logo.com/tvcountry/%s_%d.html'
# How old does the data for a region have to be before we update it (days)
web_cache_period=1
# How old does an icon have to be before we update it (days)
icon_cache_period=1
# Folder for region files
pklz_folder='regions'
# Courtesy parameter This puts delays in so as not to nark the webmaster (maybe :-| )
# You can set it to 0, but if everyone does that the server might not like us.
courtesy = 1
# # # # # # End of Program Settings # # # # # #

# Libraries required by program
import MySQLdb, difflib, heapq
from HTMLParser import HTMLParser
import urllib, urlparse
import sys, os.path, time, re, getopt,  pprint, zlib, cPickle

class lsHTMLParser(HTMLParser):
    def __init__(self, results, region_code, this_page):
        HTMLParser.__init__(self)
        self.results=results
        self.last_page=0
        self.high_page=1
        self.region_code=region_code
        self.this_page=this_page
        
    def handle_starttag(self, tag, attrs):
        if tag == 'img':
            for pair in attrs:
                if pair[0]=='src' and pair[1].find('/logo/')!=-1:
                    icon_abs_url=urlparse.urljoin(lyngsat_sub_page, pair[1])
                    icon_name=os.path.splitext(os.path.basename(icon_abs_url))[0].upper()
                    self.results[icon_name]=icon_abs_url
        # use "<A HREF="../tvcountry/uk_1.html">" to parse the highest page no
        elif tag == 'a':
            for pair in attrs:
                if pair[0]=='href' and pair[1].find('../tvcountry/%s_'%(self.region_code))!=-1:
                    link_page = int(pair[1][len('../tvcountry/%s_'%(self.region_code)):-5])
                    self.high_page=max(self.high_page, link_page)
                    if self.high_page==self.this_page:
                        self.last_page=1
                    

class RegionList:
    def __init__(self, region_codes, overwrite, force_update):
        # list of regions to test
        self.region_codes=region_codes
        self.overwrite = overwrite
        self.force_update = force_update
        self.region_list = []
        self.loadRegions()
    
    def loadRegions(self):
        for region_code in self.region_codes:
            try:
                if self.overwrite:
                    print 'Ignoring data for region "%s"...' % (region_code)
                    self.region_list.append(Region(region_code))
                # load region
                else:
                    self.region_list.append(self.loadRegion(region_code))
                    print 'Using data for region "%s"...' % (region_code)
            except:
                # otherwise we create a new object
                print 'No previous data for region "%s", creating new data...' % (region_code)
                self.region_list.append(Region(region_code))
        for region in self.region_list:
            region.grabData(self.force_update)
    
    def loadRegion(self, region_code):
        # open file, unzip and unpickle it, and append to self.region_list
        fname=os.path.join(pklz_folder, '%s.pklz' % (region_code))
        f = open(fname, 'r')
        zdata = f.read()
        f.close()
        data = zlib.decompress(zdata)
        return cPickle.loads(data)
                
    def saveRegions(self):
        if os.path.exists(pklz_folder) and not os.path.isdir(pklz_folder):
            print 'Erm... pklz_folder (%s) needs to be a folder...'
            sys.exit(2)
        if not os.path.exists(pklz_folder):
            os.mkdir(pklz_folder)
        if self.overwrite:
            # warn user and ask confirmation before destroying any previous files
            pass
        for region in self.region_list:
            self.saveRegion(region)
            
    def saveRegion(self, region):
        # pickle and gzip it
        fname=os.path.join(pklz_folder, '%s.pklz' % (region.region_code))
        data = cPickle.dumps(region)
        zdata = zlib.compress(data)
        print 'Updating/replacing "%s" region file (%s,%d,%d,%d%%)' % (
            region.region_code, fname, len(data),
            len(zdata), int((100*len(zdata))/len(data)))
        f = open(fname, 'w')
        f.write(zdata)
        f.close()
        
    def getPresets(self, search_terms):
        results=[]
        for region in self.region_list:
            results=results+region.getPresets(search_terms)
        return [[x, y, z] for x, y, z in results]
        
    def searchRegions(self, search_terms, n=10):
        results=[]
        for region in self.region_list:
            results=results+region.search(search_terms)
        uniq_results=[]
        for result in results:
            if result[3] not in [url for score, region_code, basname, url in uniq_results]:
                uniq_results.append(result)
        results = heapq.nlargest(n, uniq_results)
        return [[x, y, z] for score, x, y, z in results]

    def updateRegion(self, region_code, searchable_term, url):
        for region in self.region_list:
            if region.region_code == region_code:
                region.presets[searchable_term]=url
        
class Region:
    def __init__(self, region_code):
        #initialise structures
        self.region_code = region_code
        self.presets={}
        self.webdata={}
        self.last_webdata_update=0
        pass
        
    def grabData(self, force_update):
        # grab data into structure
        high_page='?'
        cache_expired=time.time() > (self.last_webdata_update + (web_cache_period *24*60*60))
        if force_update or cache_expired:
            # do an update rather than using cached data
            page_no = 1
            while 1:
                print 'Processing web page %d (of %s)...' % (page_no, high_page)
                while 1:
                    try:
                        opener = urllib.FancyURLopener({})
                        f = opener.open(lyngsat_sub_page % (self.region_code,page_no))
                        page_contents=f.read()
                        f.close()
                        parser=lsHTMLParser(self.webdata, self.region_code, page_no)
                        parser.feed(page_contents)
                        break
                    except:
                        pass
                    time.sleep(courtesy)
                time.sleep(courtesy)
                if parser.last_page:
                    break
                page_no = page_no+1
                high_page=parser.high_page
            self.last_webdata_update=time.time()

    def getPresets(self, search_terms):
        results=[]
        for i in range(len(search_terms)):
            term = search_terms[i]
            if self.presets.has_key(term):
                if self.presets[term] not in map(lambda x: x[2], results):
                    results.append((self.region_code, term, self.presets[term]))
        return results
        
    def search(self, search_terms):
        results=[]
        cutoff=0.3
        s=difflib.SequenceMatcher(lambda x: x == " _.")
        for i in range(len(search_terms)):
            term = search_terms[i]
            s.set_seq2(term)
            for x in self.webdata.keys():
                s.set_seq1(x)
                if s.real_quick_ratio() >= cutoff and \
                   s.quick_ratio() >= cutoff and \
                   s.ratio() >= cutoff:
                    results.append((s.ratio(), self.region_code, x, self.webdata[x]))
        return results
            
    def dumpData(self):
        pprint.pprint(self.webdata)

class ChannelList:
    def __init__(self, sourceid, force_update):
        self.force_update=force_update
        db=MySQLdb.connect(host=db_host,port=db_port,db=db_name,user=db_user,passwd=db_passwd)
        c=db.cursor()
        query = """SELECT chanid, name, callsign, xmltvid, icon FROM channel"""
        if sourceid:
            c.execute("""%s WHERE sourceid = "%s" """ % (query, sourceid))
        else:
            c.execute(query)
        inter_results=c.fetchall()
        c.close()
        self.results=[]
        for entry in inter_results:
            self.results.append([entry[0], [entry[1],entry[2],entry[3]], entry[4]])
            
    def dumpData(self):
        pprint.pprint(self.results)
        
    def retrieveIcons(self):
        channel_count=len(self.results)
        channel_current=0
        for channel_set in self.results:
            if len(channel_set)>3:
                channel_current=channel_current+1
                local_folder=os.path.join(icon_dir_local, channel_set[3][0])
                if not os.path.exists(local_folder):
                    os.makedirs(local_folder)
                local_file=os.path.join(local_folder, os.path.basename(channel_set[2]))
                if os.path.exists(local_file):
                    local_file_timestamp=os.stat(local_file)[8]
                else:
                    local_file_timestamp=0
                local_file_expired=time.time() > (local_file_timestamp + (icon_cache_period *24*60*60))
                
                remote_file=channel_set[3][2]
                if (remote_file is not None and
                   remote_file.startswith('http://') and
                   (local_file_expired or self.force_update)):
                    print 'Retrieving icon %d of %d...' % (channel_current, channel_count)
                    urllib.urlretrieve(remote_file, local_file)
            time.sleep(courtesy)
    
    def updateDBcallsign(self):
        # Note that this is very experimental. I have no way to properly test this. Assistance would be appreciated.
        db=MySQLdb.connect(host=db_host,port=db_port,db=db_name,user=db_user,passwd=db_passwd)
        c=db.cursor()
        lyngsat_common='http://www.lyngsat-logo.com/logo/tv'
        query='''INSERT INTO settings VALUES ('mythfilldatabase.urlmap.lyngsat','%s',NULL)''' % (lyngsat_common)
        for channel_set in self.results:
            if len(channel_set)>3 and channel_set[3][2].startswith('http://'):
                if channel_set[3][2].startswith(lyngsat_common):
                    url='[lyngsat]'+channel_set[3][2][len(lyngsat_common):]
                else:
                    url=channel_set[3][2]
                chan_callsign=channel_set[1][1]
                query='''insert into callsignnetworkmap (callsign,network) VALUES ('%s','lyngsat_%s')''' % (chan_callsign, chan_callsign)
                c.execute(query)
                query='''insert into networkiconmap (network,url) VALUES ('lyngsat_%s','%s')''' % (chan_callsign, url)
                c.execute(query)
        c.close()
    
    def updateDBchannel(self, sourceid):
        db=MySQLdb.connect(host=db_host,port=db_port,db=db_name,user=db_user,passwd=db_passwd)
        c=db.cursor()
        for channel_set in self.results:
            if channel_set[2] == None:
                query='''UPDATE channel SET icon="none" WHERE chanid="%s"''' % (channel_set[0])
                if sourceid is not None:
                    query = query + ''' AND sourceid=%d''' % (sourceid)
                c.execute(query)
            elif len(channel_set)>3 and channel_set[3][2].startswith("http://"):
                new_path = os.path.join(icon_dir_mythtv, channel_set[3][0], os.path.basename(channel_set[3][2]))
                query='''UPDATE channel SET icon="%s" WHERE chanid="%s"''' % (new_path, channel_set[0])
                if sourceid is not None:
                    query = query + ''' AND sourceid=%d''' % (sourceid)
                c.execute(query)
        c.close()

    def updateRegions(self, region_list, callsigns):
        for channel_set in self.results:
            if len(channel_set)>3 and channel_set[3][2].startswith('http://'):
                if callsigns:
                    searchable_term = channel_set[1][1]     # Store URL against callsign
                elif channel_set[1][2]:
                    searchable_term = channel_set[1][2]     # Store URL against xmltvid if exists
                else:
                    searchable_term = channel_set[1][0]     # Store URL against channel name
                region_list.updateRegion(channel_set[3][0], searchable_term.upper(), channel_set[3][2])
    
class Matcher:
    def __init__(self, channel_list, region_list):
        # Match icons to channels
        self.channel_list=channel_list
        self.region_list=region_list
        self.assignIcons()
        
    def assignIcons(self):
        exit=0
        autospeed=0
        channel_count=len(self.channel_list.results)
        channel_current=0
        for channel_set in self.channel_list.results:
            chan_id=channel_set[0]
            chan_name=channel_set[1][0]
            chan_callsign=channel_set[1][1]
            chan_xmltv_id=channel_set[1][2]
            chan_icon=channel_set[2]
            channel_current=channel_current+1
            search_terms=map(lambda x: x.upper(), channel_set[1])
            while exit==0:
                headerstring=80*'-'
                headerstring_offset=2
                headerstring_left='Channel ID: %s' % (chan_id)
                headerstring_right='[%03d/%03d]' % (channel_current, channel_count)
                headerstring=(headerstring[0:headerstring_offset]+
                                    headerstring_left+
                                    headerstring[len(headerstring_left)+headerstring_offset:-(len(headerstring_right)+headerstring_offset)]+
                                    headerstring_right+
                                    headerstring[-headerstring_offset:])
                print '\n\n'+headerstring
                print '        Name: %s' % (chan_name)
                print '    Callsign: %s' % (chan_callsign)
                print '    XMLTV ID: %s' % (chan_xmltv_id)
                print '   Icon Path: %s' % (chan_icon)
                print
                # Find exact matches in previous entries
                presets=self.region_list.getPresets(search_terms)
                if not len(presets):
                    print 'I have no matching presets.'
                elif len(presets)==1:
                    print 'Found one exact matching preset for %s, %s' % (presets[0][0], presets[0][1])
                else:
                    print 'Found multiple matching presets. Selecting will give menu.'

                # Find fuzzy matches
                option_list=self.region_list.searchRegions(search_terms)
                if len(option_list)>0:
                    print 'The best fuzzy matches (up to ten) are:'
                    num_prompt='1'
                    if len(option_list)>1:
                        num_prompt=num_prompt+'-%d' % len(option_list)
                    num_prompt=num_prompt+'/'
                else:
                    print 'No suitable fuzzy matches found!'
                    num_prompt = ''
                # build menu
                for i in range(0, len(option_list)):
                    entry='    %2d. %s, %-20s (%s)' % (i+1,  option_list[i][0],
                                                                            option_list[i][1],
                                                                            option_list[i][2][len('http://www.lyngsat-logo.com/logo/tv'):])[:80]
                    print entry
                # get user response
                num_resp_options=[]
                for entry in range(1,len(option_list)+1):
                    num_resp_options.append(str(entry))
                resp_options=(len(presets) and ['c','f','p','a','.',''] or ['c','f','a','.',''])+num_resp_options
                resp=None
                if autospeed:
                    if chan_icon=='none':
                        if len(presets):
                            resp='p'
                        else:
                            autospeed=0
                while resp==None:
                    print """Enter a number to select icon from list, "f" to enter a fuzzy term,
"a" to %s autospeed, "c" to clear, %s
"." to skip all remaining channels or <Return> to skip this channel.""" % (
                                            autospeed and 'stop' or 'start',
                                            len(presets) and '"p" to select preset list entry,' or '')
                    resp=raw_input('Please enter choice [%s%sc/f/a/s/./<Return>]? ' % (
                                            num_prompt, len(presets) and 'p/' or ''))
                    if resp not in resp_options:
                        resp=None
                        print
                if resp in num_resp_options:
                    channel_set.append(option_list[int(resp)-1])
                    break
                if resp == 'a':
                    if autospeed:
                        autospeed=0
                    else:
                        autospeed=1
                if resp == 'p':
                    if len(presets) == 1:
                        channel_set.append(presets[0])
                        break
                    else:
                        # get choice from user
                        print '\n\n[%d/%d] Channel id: %s is currently set to "%s".' % (channel_current, channel_count, chan_id, chan_icon)
                        print 'MythTV identities=     Name: %s' % (chan_name)
                        print '                   Callsign: %s' % (chan_callsign)
                        print '                   XMLTV ID: %s' % (chan_xmltv_id)
                        print 'The available presets are:'
                        for i in range(0, len(presets)):
                            print '    %2d. %s, %s' % (i+1, presets[i][0], presets[i][0])
                        num_resp_options=[]
                        for entry in range(1,len(presets)+1):
                            num_resp_options.append(str(entry))
                        resp_options=['r']+num_resp_options
                        resp=None
                        while resp==None:
                            print """Enter a number to select icon from list, or "r" to return to main menu."""
                            resp=raw_input('Please enter choice [1-%d/r]? ' % (len(presets)))
                            if resp not in resp_options:
                                resp=None
                                print
                            if resp in num_resp_options:
                                channel_set.append(presets[int(resp)-1])
                                ret_main=0
                            if resp =='r':
                                ret_main=1
                        if not ret_main:
                            break
                if resp == 'c':
                    channel_set[2]=None
                    break
                if resp == 'f':
                    search_term=raw_input('Enter a new fuzzy term: ')
                    search_terms=[search_term.upper()]
                if resp == '':
                    break
                if resp == '.':
                    print
                    exit=1
                    break
        
def usage():
    print __doc__

def main():
    # extract arguments
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'hr:s:oc',
            ['help', 'regions=', 'source=', 'overwrite', 'callsign'])
    except getopt.GetoptError:
        # print help information and exit:
        print 'GetOpt Error'
        usage()
        sys.exit(2)
    
    # Default settings
    region_codes='uk'
    sourceid=None
    overwrite=0
    callsigns=0
    force_update=0
    
    # Process cmd line options
    for o, a in opts:
        if o in ('-h', '--help'):
            usage()
            sys.exit()
        if o in ('-r', '--regions'):
            region_codes = a
        if o in ('-s', '--source'):
            sourceid = a
        if o in ('-o', '--overwrite'):
            overwrite = 1
        if o in ('-c', '--callsigns'):
            callsigns = 1
        if o in ('-f', '--force-update'):
            force_update = 1
    
    # Open the previous data files
    region_codes = region_codes.split(',')
    region_list=RegionList(region_codes, overwrite, force_update)
    
    # Get DB channels
    channel_list=ChannelList(sourceid, force_update)
    #channel_list.dumpData()
    
    # Assign icons
    matcher=Matcher(channel_list, region_list)
    
    # The useful bit :-)
    if callsigns:
        # Update DB callsignnetworkmap, networkiconmap, settings
        channel_list.updateDBcallsign()
    else:
        # Retrieve icons
        channel_list.retrieveIcons()
        # Update DB channel
        channel_list.updateDBchannel(sourceid)
    
    # Update data held for the regions
    channel_list.updateRegions(region_list, callsigns)
    
    # Write our updates to the local data files
    region_list.saveRegions()
    
if __name__=='__main__':
    main()
    sys.exit(0)
